from .gdsast import *
from .geoms import Vector2D

def _gds_element_flip_rotate_pt(el, pt):
    if "mag" in el and el["element"] == "sref":
        raise NotImplemented
    flip = "strans" in el and el["strans"]&0x8000 # flipped
    angle = el.get("angle", 0)
    return gds_flip_rotate_pt(pt, flip, angle)

def _gds_element_update_pt(el, pt):
    pt = _gds_element_flip_rotate_pt(el, pt)
    x, y = el["xy"][0]
    return pt[0] + x, pt[1] + y

def _flip_rotate_orient(orient, flip, angle):
    if flip:
        if orient == "N":
            orient = 'S'
        elif orient == 'S':
            orient = 'N'
    if not angle:
        pass
    elif angle == 90:
        orient = {'N':'W','W':'S','S':'E','E':'N'}[orient]
    elif angle == 180:
        orient = {'N':'S','W':'E','S':'N','E':'W'}[orient]
    elif angle == 270:
        orient = {'N':'E','W':'N','S':'W','E':'S'}[orient]
    else:
        assert(False),"Expecting angle be 0, 90, 180 or 270"
    return orient


def _flip_rotate_pins(pins, flip, angle):
    _pins = {}
    for _orient, labs in pins.items():
        _labs = _pins[_flip_rotate_orient(_orient, flip, angle)] = {}
        for label, info in labs.items():
            _info = _labs[label] = {}
            for layer, ptsw in info.items():
                _ptsw = _info[layer] = {(gds_flip_rotate_pt(ptw[0], flip, angle),ptw[1]) for ptw in ptsw}
    return _pins


def _point_inside_geom(_geom, pt):
    v = Vector2D(pt[0], pt[1])
    if _geom["element"] == "boundary":
        return v.inside(_geom["xy"])
    elif _geom["element"] == "path":
        width = _geom["width"]
        def _boundary_pts():
            back = []
            def _ofs(pt1, pt2):
                dv = Vector2D(pt2[0] - pt1[0], pt2[1] - pt1[1]).normal().sized(width / 2)
                dv.x = int(round(dv.x))
                dv.y = int(round(dv.y))
                yield pt1 + dv
                yield pt2 + dv
                back.insert(0, pt1 - dv)
                back.insert(0, pt2 - dv)
            prev = None
            for pt in _geom["xy"]:
                if prev is not None:
                    yield from _ofs(prev, pt)
                prev = pt
            yield from back
        return v.inside(_boundary_pts())

def _box_pts(pts):
    xy = list(zip(*pts))    # -> [[all Xs], [all Ys]]
    if xy:
        yield list(map(min,xy)) # [min of all Xs, min of all Ys]
        yield list(map(max,xy)) # [max of all Xs, max of all Ys]

class GDSLayout:
    def __init__(self):
        self.gds = None # AST
        self.sts = {}   # structures map by name

    def init(self, lib_name, units = None):
        self.sts = {}
        if not units:
            units = [0.001, 1e-9]
        self.gds = gds_create(lib_name, units)

    def u2units(self, microns): # return GDS units from micrometers (1e-9 meters)
        units = self.gds["units"]
        return int(round(units[1] / units[0] / 1e-9 * microns))

    def units2u(self, units): # return microns (1e-9 meters) from GDS units
        _units = self.gds["units"]
        return units * _units[0] / (_units[1] * 1e+9)

    def load_file(self, filename):
        with open(filename,"rb") as f:
            _gds = gds_read(f)
            if self.gds is None:
                self.gds = _gds
                for st in _gds["structures"]:
                    self.sts[st["name"]] = st
            else:
                units = self.gds["units"]
                _units = _gds["units"]
                if units != _units:
                    raise NotImplemented
                for st in _gds["structures"]:
                    name = st["name"]
                    if name in self.sts:
                        # TODO: check is same
                        print(f"GDS is reusing structure '{name}' while loading '{filename}'")
                        continue
                    self.sts[name] = st
                    self.gds["structures"].append(st)

    def get_structure(self, name):
        return self.sts.get(name)

    def new_structure(self, name):
        st = self.get_structure(name)
        if st is not None:
            self.gds["structures"].remove(st)
        st = gds_create_structure(name)
        self.gds["structures"].append(st)
        self.sts[name] = st
        return st

    def rename_structure(self, from_name, to_name):
        if from_name == to_name:
            return
        assert(to_name not in self.sts),f"Can't rename structure '{from_name}' to '{to_name}' as name been used"
        st = self.get_structure(from_name)
        self.sts[from_name]
        st["name"] = to_name
        self.sts[to_name] = st

    @staticmethod
    def _is_el_match(el, filt):
        if not filt:
            return False
        def _match(name, val):
            if name in el:
                if isinstance(val, (set, tuple, list)):
                    return el[name] in val
                else:
                    return el[name] == val
            return False
        if isinstance(filt, dict):
            for name, _filt in filt.items():
                if _match(name, _filt):
                    return True
        else:
            return _match("element", filt)

    def get_structure_geom_points(self, st, ignore = "text"):
        for el in st["elements"]:
            yield from self.get_element_geom_points(el, ignore)

    def get_element_geom_points(self, el, ignore = "text"):
        if not self._is_el_match(el, ignore):
            if el["element"] == "sref":
                if "mag" in el:
                    raise NotImplemented
                x, y = el["xy"][0] # sref position
                for pt in self.get_structure_geom_points(el["name"]):
                    pt = _gds_element_flip_rotate_pt(el, pt)
                    yield pt[0] + x, pt[1] + y
            elif "xy" in el:
                yield from el["xy"]

    def get_structure_geom_box(self, name, ignore = "text"):
        st = self.sts[name]
        def _pts():
            for el in st["elements"]:
                yield from self.get_element_geom_box(el, ignore)
        return list(_box_pts(_pts()))

    def get_element_geom_box(self, el, ignore = "text"):
        def _pts():
            if not self._is_el_match(el, ignore) and "xy" in el:
                xy = el["xy"]
                if el["element"] == "sref":
                    if "mag" in el:
                        raise NotImplemented
                    x, y = xy[0] # sref position
                    for pt in self.get_structure_geom_box(el["name"], ignore):
                        pt = _gds_element_flip_rotate_pt(el, pt)
                        yield pt[0] + x, pt[1] + y
                else:
                    yield from _box_pts(xy)
        return list(_box_pts(_pts()))

    def _ordered_names_with_deps(self, st_names):
        names = set()
        def _srefs(st):
            for el in st["elements"]:
                if el["element"] == "sref":
                    yield el
        def _populate(name):
            names.add(name)
            st = self.get_structure(name)
            assert(st), f"GDS referencing non existing structure '{name}"
            for el in _srefs(st):
                el_name = el["name"]
                if el_name in names:
                    continue
                _populate(el_name)
        for name in st_names:
            _populate(name)
        # dependancy sort
        ordered_names = sorted(names)
        i = 0
        nerr = 0
        while i<len(ordered_names):
            name = ordered_names[i]
            st = self.get_structure(name)
            if any(map(lambda el: el["name"] in  names, _srefs(st))):
                ordered_names.pop(i)
                ordered_names.append(name)
                nerr += 1
                assert(nerr < len(names)), "GDS structures have elements with cross dependancies"
            else:
                names.discard(name)
                i += 1
                nerr = 0
        return ordered_names

    def new_gds(self, lib_name, st_names): # create new GDSLayout with only selected structures
        gds = GDSLayout()
        gds.init(lib_name, self.gds["units"])
        for name in self._ordered_names_with_deps(st_names):
            st = self.get_structure(name)
            gds.gds["structures"].append(st)
        return gds

    def save_structures(self, filename, lib_name, st_names):
        gds = self.new_gds(lib_name, st_names)
        gds.save_file(filename)

    def save_file(self, filename):
        with open(filename, "wb") as f:
            gds_write(f, self.gds)

    def extract_pins(self, cell_name, ignore = None):
        st = self.get_structure(cell_name)
        assert(st),f"Unknown '{cell_name}' cell"
        def _labels(st):
            for el in st["elements"]:
                if el["element"] == "text":
                    yield el["xy"][0], el["layer"], el["text"]
                elif el["element"] == "sref":
                    for pt, layer, text in _labels(self.get_structure(el["name"])):
                        yield _gds_element_update_pt(el, pt), layer, text
        layer2labs = {}
        for pt, layer, label in _labels(st):
            if layer in layer2labs:
                labs = layer2labs[layer]
            else:
                labs = layer2labs[layer] = {}
            if label in labs:
                pts = labs[label]
            else:
                pts = labs[label] = set()
            pts.add(pt)
        pmin, pmax = self.get_structure_geom_box(cell_name, ignore)
        def _update_xy(sref, xy):
            for pt in xy:
                yield _gds_element_update_pt(sref, pt)
        def _geoms(st): # all geometries in the structure
            for el in st["elements"]:
                el_type = el["element"]
                if el_type == "sref":
                    for geom in _geoms(self.get_structure(el["name"])):
                        _geom = geom.copy()
                        _geom["xy"] = list(_update_xy(el, geom["xy"]))
                        yield _geom
                if el_type in ("boundary", "path"):
                    yield el
        def _edge_geoms(): # only geometries that touch edges
            for _geom in _geoms(st):
                for pt in _geom["xy"]:
                    if (pt[0] == pmin[0] or
                        pt[1] == pmin[1] or
                        pt[0] == pmax[0] or
                        pt[1] == pmax[1]):
                        yield _geom
                        break
        def _find_geom_label(_geom):
            layer = _geom["layer"]
            if layer in layer2labs:
                for label, pts in layer2labs[layer].items():
                    for pt in pts:
                        if _point_inside_geom(_geom,pt):
                            return label
        def _labeled_edge_geoms(): # all geometries that have label inside them
            for _geom in _edge_geoms():
                label = _find_geom_label(_geom)
                if label:
                    yield label, _geom
        pins = {} # map chain: orient->label->layer->(pt,width) 
        def _add_pin(orient, label, layer, pt, width):
            if orient in pins:
                side_pins = pins[orient]
            else:
                side_pins = pins[orient] = {}
            key = (label, layer)
            if label in side_pins:
                lmap = side_pins[label]
            else:
                lmap = side_pins[label] = {}
            if layer in lmap:
                pts = lmap[layer]
            else:
                pts = lmap[layer] = set()
            pts.add((pt, width))
        for label, _geom in _labeled_edge_geoms():
            if _geom["element"] == "path":
                width = _geom["width"]
                layer = _geom["layer"]
                for pt in _geom["xy"]:
                    if pt[0] == pmin[0]:
                        _add_pin('W', label, layer, pt, width)
                    elif pt[0] == pmax[0]:
                        _add_pin('E', label, layer, pt, width)
                    elif pt[1] == pmin[1]:
                        _add_pin('S', label, layer, pt, width)
                    elif pt[1] == pmax[1]:
                        _add_pin('N', label, layer, pt, width)
            elif _geom["element"] == "boundary":
                layer = _geom["layer"]
                prev = None
                for pt in _geom["xy"]:
                    if prev is not None:
                        if prev[0] == pt[0]:
                            if pt[0] == pmin[0]:
                                _add_pin('W', label, layer, (pt[0], (prev[1]+pt[1]+1)//2), abs(pt[1] - prev[1]))
                            elif pt[0] == pmax[0]:
                                _add_pin('E', label, layer, (pt[0], (prev[1]+pt[1]+1)//2), abs(pt[1] - prev[1]))
                        elif prev[1] == pt[1]:
                            if pt[1] == pmin[1]:
                                _add_pin('S', label, layer, ((prev[0]+pt[0]+1)//2, pt[1]), abs(pt[0] - prev[0]))
                            elif pt[1] == pmax[1]:
                                _add_pin('N', label, layer, ((prev[0]+pt[0]+1)//2, pt[1]), abs(pt[0] - prev[0]))
                    prev = pt
        return pins  # map chain: orient->label->layer->[(pt,width)] of edge points of labeled geometries

    def _sref_pins(self, sref, pins):
        flip = "strans" in sref and sref["strans"]&0x8000 # flipped
        angle = sref.get("angle", 0)
        return _flip_rotate_pins(pins, flip, angle)

# helper class to work with GDS structure existing or new
class GDSStructure:
    def __init__(self, gds, name):
        self.gds = gds
        st = gds.get_structure(name)
        if st is None:
            st = gds.new_structure(name)
        self.items = st["elements"]
        self.srefs = {}

    def add_sref(self, sref_name, ref_name, flipped, angle):
        assert(ref_name in self.gds.sts),f"Referencing unknown '{ref_name}' structure"
        item = {
            "element" : "sref",
            "name" : ref_name,
#            "sref_name" : sref_name
            "xy" : [(0, 0)]
            }
        if angle or flipped:
            item["strans"] = 0x8000 if flipped else 0
        if angle:
            item["angle"] = float(angle)
        self.items.append(item)
        self.srefs[sref_name] = item
        return item

    def add_path(self, pt1, pt2, width, layer):
        item = {
            "element" : "path",
            "layer": int(layer),
            "datatype": 0,
            "width": int(width),
            "xy" : [tuple(pt1), tuple(pt2)]
            }
        self.items.append(item)
        return item

    def add_rect(self, box, layer):
        def _pts():
            x_min = min(box[0][0], box[1][0])
            x_max = max(box[0][0], box[1][0])
            y_min = min(box[0][1], box[1][1])
            y_max = max(box[0][1], box[1][1])
            yield (x_min, y_min)
            yield (x_max, y_min)
            yield (x_max, y_max)
            yield (x_min, y_max)
            yield (x_min, y_min)
        item = {
            "element" : "boundary",
            "layer": int(layer),
            "datatype": 0,
            "xy" : list(_pts())
            }
        self.items.append(item)
        return item

    #def add_via(self, box, size, layer, spacing = None, nx = None, ny = None, grid = None):
        #if not spacing:
            #spacing = size
        #width = box[1][0] - box[0][0]
        #height = box[1][1] - box[0][1]
        #if not nx:
            #nx = (width - spacing) // (size + spacing)
        #if not ny:
            #ny = (height - spacing) // (size + spacing)
        #xofs = box[0][0] + (width - nx * (size + spacing) + spacing) // 2
        #yofs = box[0][1] + (height - ny * (size + spacing) + spacing) // 2
        #if grid:
            #def _grid_update(v):
                #diff = v % grid
                #if diff *2 < grid:
                    #diff = -diff
                #else:
                    #diff = grid - diff
                #return v + diff
            #xofs = _grid_update(xofs)
            #yofs = _grid_update(yofs)
        ##print(f"VIA: x={xofs} y={yofs} size={size} nx={nx} ny={ny} width={width} height={height}")
        #for j in range(ny):
            #for i in range(nx):
                #x = xofs + i * (size + spacing)
                #y = yofs + j * (size + spacing) + size // 2
                #self.add_path((x, y), (x + size, y), size, layer)



import struct
import time
import math

GDS_NO_DATA, GDS_BIT_ARRAY, GDS_INT16, GDS_INT32, GDS_REAL32, GDS_REAL64, GDS_STR = range(7)

def GDS_REC16(rec_id, gds_type):
    return (rec_id<<8)|gds_type

GDS_HEADER       = GDS_REC16(0x00, GDS_INT16)
GDS_BGNLIB       = GDS_REC16(0x01, GDS_INT16)
GDS_LIBNAME      = GDS_REC16(0x02, GDS_STR)
GDS_UNITS        = GDS_REC16(0x03, GDS_REAL64)
GDS_ENDLIB       = GDS_REC16(0x04, GDS_NO_DATA)
GDS_BGNSTR       = GDS_REC16(0x05, GDS_INT16)
GDS_STRNAME      = GDS_REC16(0x06, GDS_STR)
GDS_ENDSTR       = GDS_REC16(0x07, GDS_NO_DATA)
GDS_BOUNDARY     = GDS_REC16(0x08, GDS_NO_DATA)
GDS_PATH         = GDS_REC16(0x09, GDS_NO_DATA)
GDS_SREF         = GDS_REC16(0x0A, GDS_NO_DATA)
GDS_AREF         = GDS_REC16(0x0B, GDS_NO_DATA)
GDS_TEXT         = GDS_REC16(0x0C, GDS_NO_DATA)
GDS_LAYER        = GDS_REC16(0x0D, GDS_INT16)
GDS_DATATYPE     = GDS_REC16(0x0E, GDS_INT16)
GDS_WIDTH        = GDS_REC16(0x0F, GDS_INT32)
GDS_XY           = GDS_REC16(0x10, GDS_INT32)
GDS_ENDEL        = GDS_REC16(0x11, GDS_NO_DATA)
GDS_SNAME        = GDS_REC16(0x12, GDS_STR)
GDS_COLROW       = GDS_REC16(0x13, GDS_INT16)
GDS_TEXTNODE     = GDS_REC16(0x14, GDS_NO_DATA)
GDS_NODE         = GDS_REC16(0x15, GDS_NO_DATA)
GDS_TEXTTYPE     = GDS_REC16(0x16, GDS_INT16)
GDS_PRESENTATION = GDS_REC16(0x17, GDS_BIT_ARRAY)
GDS_STRING       = GDS_REC16(0x19, GDS_STR)
GDS_STRANS       = GDS_REC16(0x1A, GDS_BIT_ARRAY)
GDS_MAG          = GDS_REC16(0x1B, GDS_REAL64)
GDS_ANGLE        = GDS_REC16(0x1C, GDS_REAL64)
GDS_REFLIBS      = GDS_REC16(0x1F, GDS_STR)
GDS_FONTS        = GDS_REC16(0x20, GDS_STR)
GDS_PATHTYPE     = GDS_REC16(0x21, GDS_INT16)
GDS_GENERATIONS  = GDS_REC16(0x22, GDS_INT16)
GDS_ATTRTABLE    = GDS_REC16(0x23, GDS_STR)
GDS_ELFLAGS      = GDS_REC16(0x26, GDS_BIT_ARRAY)
GDS_NODETYPE     = GDS_REC16(0x2A, GDS_INT16)
GDS_PROPATTR     = GDS_REC16(0x2B, GDS_INT16)
GDS_PROPVALUE    = GDS_REC16(0x2C, GDS_STR)
GDS_BOX          = GDS_REC16(0x2D, GDS_NO_DATA)
GDS_BOXTYPE      = GDS_REC16(0x2E, GDS_INT16)
GDS_PLEX         = GDS_REC16(0x2F, GDS_INT32)
GDS_BGNEXTN      = GDS_REC16(0x30, GDS_INT32)
GDS_ENDEXTN      = GDS_REC16(0x31, GDS_INT32)
GDS_TAPENUM      = GDS_REC16(0x32, GDS_INT16)
GDS_TAPECODE     = GDS_REC16(0x33, GDS_INT16)
GDS_STRCLASS     = GDS_REC16(0x34, GDS_BIT_ARRAY)
GDS_FORMAT       = GDS_REC16(0x36, GDS_INT16)
GDS_MASK         = GDS_REC16(0x37, GDS_STR)
GDS_ENDMASKS     = GDS_REC16(0x38, GDS_NO_DATA)

GDS_VALID_TAGS = {
    GDS_HEADER,
    GDS_BGNLIB,
    GDS_LIBNAME,
    GDS_UNITS,
    GDS_ENDLIB,
    GDS_BGNSTR,
    GDS_STRNAME,
    GDS_ENDSTR,
    GDS_BOUNDARY,
    GDS_PATH,
    GDS_SREF,
    GDS_AREF,
    GDS_TEXT,
    GDS_LAYER,
    GDS_DATATYPE,
    GDS_WIDTH,
    GDS_XY,
    GDS_ENDEL,
    GDS_SNAME,
    GDS_COLROW,
    GDS_TEXTNODE,
    GDS_NODE,
    GDS_TEXTTYPE,
    GDS_PRESENTATION,
    GDS_STRING,
    GDS_STRANS,
    GDS_MAG,
    GDS_ANGLE,
    GDS_REFLIBS,
    GDS_FONTS,
    GDS_PATHTYPE,
    GDS_GENERATIONS,
    GDS_ATTRTABLE,
    GDS_ELFLAGS,
    GDS_NODETYPE,
    GDS_PROPATTR,
    GDS_PROPVALUE,
    GDS_BOX,
    GDS_BOXTYPE,
    GDS_PLEX,
    GDS_BGNEXTN,
    GDS_ENDEXTN,
    GDS_TAPENUM,
    GDS_TAPECODE,
    GDS_STRCLASS,
    GDS_FORMAT,
    GDS_MASK,
    GDS_ENDMASKS,
    }

def _gds_unpack_no_data(data):
    return None

def _gds_pack_no_data(v):
    return b''

def _gds_unpack_bit_array(data):
    return data

def _gds_pack_bit_array(v):
    return v

def _gds_unpack_int16(data):
    nums = list(map(lambda v: v[0],struct.iter_unpack(">h",data)))
    if len(nums)==1:
        return nums[0]
    return nums

def _gds_pack_int16(v):
    if isinstance(v,int):
        return struct.pack('>h',v)
    data = b''
    for _v in v:
        data += struct.pack('>h',_v)
    return data

def _gds_unpack_uint16(data):
    nums = list(map(lambda v: v[0],struct.iter_unpack(">H",data)))
    if len(nums)==1:
        return nums[0]
    return nums

def _gds_pack_uint16(v):
    if isinstance(v,int):
        return struct.pack('>H',v)
    data = b''
    for _v in v:
        data += struct.pack('>H',_v)
    return data

def _gds_unpack_int32(data):
    nums = list(map(lambda v: v[0],struct.iter_unpack(">i",data)))
    if len(nums)==1:
        return nums[0]
    return nums

def _gds_pack_int32(v):
    if isinstance(v,int):
        return struct.pack('>i',v)
    data = b''
    for _v in v:
        data += struct.pack('>i',_v)
    return data

def _gds_unpack_real32(data):
    assert((len(data)&3)==0),"Expecting multiple of 4 bytes data"
    def _nums():
        for i in range(0,len(data),4):
            b1 = data[i]
            v32 = struct.unpack(">I",b'\0'+data[i+1:i+4])[0]
            v = math.ldexp(v32, ((b1&0x7F)*4)-280)
            if b1 & 0x80:
                yield -v
            else:
                yield v
    nums = list(_nums())
    if len(nums)==1:
        return nums[0]
    return nums

def _gds_pack_real32(v):
    def _pack(_v):
        if _v<0:
            b1 = 0x80
            _v = -_v
        else:
            b1 = 0
        m, exp = math.frexp(_v)
        m = int(m * math.pow(2, 24))
        exp -= 24
        while exp&3:
            exp += 1
            m >>= 1
        exp = (exp + 280)>>2
        data = struct.pack('>I', m)
        return struct.pack('B', b1|(exp&0x7F))+data[1:]
    if isinstance(v, (int, str)):
        v = float(v)
    if isinstance(v, float):
        return _pack(v)
    data = b''
    for _v in v:
        data += _pack(_v)
    raise data

def _gds_unpack_real64(data):
    assert((len(data)&7)==0),"Expecting multiple of 8 bytes data"
    def _nums():
        for i in range(0,len(data),8):
            b1 = data[i]
            v64 = struct.unpack(">Q",b'\0'+data[i+1:i+8])[0]
            v = math.ldexp(v64, ((b1&0x7F)*4)-312)
            if b1 & 0x80:
                yield -v
            else:
                yield v
    nums = list(_nums())
    if len(nums)==1:
        return nums[0]
    return nums

def _gds_pack_real64(v):
    def _pack(_v):
        if _v<0:
            b1 = 0x80
            _v = -_v
        else:
            b1 = 0
        m, exp = math.frexp(_v)
        m = int(m * math.pow(2, 56))
        exp -= 56
        while exp&3:
            exp += 1
            m >>= 1
        exp = (exp + 312)>>2
        data = struct.pack('>Q', m)
        return struct.pack('B', b1|(exp&0x7F))+data[1:]
    if isinstance(v, (int, str)):
        v = float(v)
    if isinstance(v, float):
        return _pack(v)
    data = b''
    for _v in v:
        data += _pack(_v)
    return data

def _gds_unpack_str(data):
    i = data.find(b'\0')
    if i>=0:
        data = data[:i]
    return data.decode('utf-8')

def _gds_pack_str(v):
    if isinstance(v, str):
        data = v.encode('utf-8')
        if len(data)&1:
            data += b'\0'
    else:
        raise NotImplemented
    return data

def _gds_unpack_xy(data):
    def _pts():
        xy = iter(_gds_unpack_int32(data))
        while True:
            try:
                x = next(xy)
                y = next(xy)
                yield (x, y)
            except StopIteration:
                break
    return list(_pts())

def _gds_pack_xy(v):
    def _xy():
        for pt in v:
            yield pt[0]
            yield pt[1]
    return _gds_pack_int32(_xy())

_gds_unpack_map = {
    GDS_NO_DATA   : _gds_unpack_no_data,
    GDS_BIT_ARRAY : _gds_unpack_uint16,
    GDS_INT16     : _gds_unpack_int16,
    GDS_INT32     : _gds_unpack_int32,
    GDS_REAL32    : _gds_unpack_real32,
    GDS_REAL64    : _gds_unpack_real64,
    GDS_STR       : _gds_unpack_str,
    GDS_XY        : _gds_unpack_xy
}

_gds_pack_map = {
    GDS_NO_DATA   : _gds_pack_no_data,
    GDS_BIT_ARRAY : _gds_pack_uint16,
    GDS_INT16     : _gds_pack_int16,
    GDS_INT32     : _gds_pack_int32,
    GDS_REAL32    : _gds_pack_real32,
    GDS_REAL64    : _gds_pack_real64,
    GDS_STR       : _gds_pack_str,
    GDS_XY        : _gds_pack_xy
}

def _gds_unpack_data(rec_id, data):
    unpack_proc = _gds_unpack_map.get(rec_id)
    if unpack_proc is None:
        unpack_proc = _gds_unpack_map[rec_id&0xFF]
    return unpack_proc(data)

def _gds_pack_data(rec_id, value):
    pack_proc = _gds_pack_map.get(rec_id)
    if pack_proc is None:
        pack_proc = _gds_pack_map[rec_id&0xFF]
    return pack_proc(value)

GDS_OPTIONAL   = 0x800000
GDS_NON_TOKEN  = 0x010000
GDS_ARRAY      = 0x020000
GDS_INCLUDE    = 0x040000
GDS_CHOICE     = 0x080000
GDS_TAGMASK    = 0x01FFFF
GDS_NONTAGMASK = 0xFF0000
GDS_LIBRARY       =  0|GDS_NON_TOKEN
GDS_FORMATTYPE    =  1|GDS_NON_TOKEN
GDS_STRUCTURE     =  2|GDS_NON_TOKEN
GDS_MASKS         =  3|GDS_NON_TOKEN
GDS_ELEMENT       =  4|GDS_NON_TOKEN
GDS_PROPERTY      =  5|GDS_NON_TOKEN
GDS_BOUNDARY_ELEM =  6|GDS_NON_TOKEN
GDS_PATH_ELEM     =  7|GDS_NON_TOKEN
GDS_SREF_ELEM     =  8|GDS_NON_TOKEN
GDS_AREF_ELEM     =  9|GDS_NON_TOKEN
GDS_TEXT_ELEM     = 10|GDS_NON_TOKEN
GDS_NODE_ELEM     = 11|GDS_NON_TOKEN
GDS_BOX_ELEM      = 12|GDS_NON_TOKEN
GDS_STRANS_GROUP  = 13|GDS_NON_TOKEN
GDS_TEXTBODY      = 14|GDS_NON_TOKEN

_gds_parse_schema = {
    GDS_LIBRARY: [
        (GDS_HEADER,"version"),
        (GDS_BGNLIB,"timestamp"),
        (GDS_LIBNAME,'name'),
        (GDS_REFLIBS|GDS_OPTIONAL,"reflibs"),
        (GDS_FONTS|GDS_OPTIONAL,"fonts"),
        (GDS_ATTRTABLE|GDS_OPTIONAL,"attributes"),
        #(GDS_STYPTABLE|GDS_OPTIONAL,""),
        (GDS_GENERATIONS|GDS_OPTIONAL,"gens"),
        (GDS_FORMATTYPE|GDS_OPTIONAL,"format_type"),
        (GDS_UNITS,"units"),
        (GDS_STRUCTURE|GDS_OPTIONAL|GDS_ARRAY,"structures"),
        (GDS_ENDLIB,None)
        ],
    GDS_FORMATTYPE: [
        (GDS_FORMAT,"format"),
        (GDS_MASKS|GDS_INCLUDE|GDS_OPTIONAL,None),
        ],
    GDS_MASKS: [
        (GDS_MASK|GDS_ARRAY,"masks"),
        (GDS_ENDMASKS,None)
        ],
    GDS_STRUCTURE: [
        (GDS_BGNSTR,"timestamp"),
        (GDS_STRNAME,"name"),
        (GDS_STRCLASS|GDS_OPTIONAL,"class"),
        #(GDS_STRTYPE|GDS_OPTIONAL,"type"),
        (GDS_ELEMENT|GDS_ARRAY|GDS_OPTIONAL,"elements"),
        (GDS_ENDSTR,None)
        ],
    GDS_ELEMENT: [
        (GDS_CHOICE, [
            GDS_BOUNDARY_ELEM,
            GDS_PATH_ELEM,
            GDS_SREF_ELEM,
            GDS_AREF_ELEM,
            GDS_TEXT_ELEM,
            GDS_NODE_ELEM,
            GDS_BOX_ELEM]),
        #(GDS_ELKEY|GDS_OPTIONAL,"key"),
        (GDS_PROPERTY|GDS_OPTIONAL|GDS_ARRAY,"properties"),
        (GDS_ENDEL,None)
        ],
    GDS_BOUNDARY_ELEM: [
        (GDS_BOUNDARY,{"element": "boundary"}),
        (GDS_ELFLAGS|GDS_OPTIONAL,"flags"),
        (GDS_PLEX|GDS_OPTIONAL,"plex"),
        (GDS_LAYER,"layer"),
        (GDS_DATATYPE,"datatype"),
        (GDS_XY,"xy")
        ],
    GDS_PATH_ELEM: [
        (GDS_PATH,{"element": "path"}),
        (GDS_ELFLAGS|GDS_OPTIONAL,"flags"),
        (GDS_PLEX|GDS_OPTIONAL,"plex"),
        (GDS_LAYER,"layer"),
        (GDS_DATATYPE,"datatype"),
        (GDS_PATHTYPE|GDS_OPTIONAL,"pathtype"),
        (GDS_WIDTH|GDS_OPTIONAL,"width"),
        (GDS_BGNEXTN|GDS_OPTIONAL,"begin_ext"),
        (GDS_ENDEXTN|GDS_OPTIONAL,"end_ext"),
        (GDS_XY,"xy")
        ],
    GDS_SREF_ELEM: [
        (GDS_SREF,{"element": "sref"}),
        (GDS_ELFLAGS|GDS_OPTIONAL,"flags"),
        (GDS_PLEX|GDS_OPTIONAL,"plex"),
        (GDS_SNAME,"name"),
        (GDS_STRANS_GROUP|GDS_OPTIONAL|GDS_INCLUDE,None),
        (GDS_XY,"xy")
        ],
    GDS_AREF_ELEM: [
        (GDS_AREF,{"element": "aref"}),
        (GDS_ELFLAGS|GDS_OPTIONAL,"flags"),
        (GDS_PLEX|GDS_OPTIONAL,"plex"),
        (GDS_SNAME,"name"),
        (GDS_STRANS_GROUP|GDS_OPTIONAL|GDS_INCLUDE,None),
        (GDS_COLROW,"colrow"),
        (GDS_XY,"xy")
        ],
    GDS_TEXT_ELEM: [
        (GDS_TEXT,{"element": "text"}),
        (GDS_ELFLAGS|GDS_OPTIONAL,"flags"),
        (GDS_PLEX|GDS_OPTIONAL,"plex"),
        (GDS_LAYER,"layer"),
        (GDS_TEXTBODY|GDS_INCLUDE,None)
        ],
    GDS_NODE_ELEM: [
        (GDS_NODE,{"element": "node"}),
        (GDS_ELFLAGS|GDS_OPTIONAL,"flags"),
        (GDS_PLEX|GDS_OPTIONAL,"plex"),
        (GDS_LAYER,"layer"),
        (GDS_NODETYPE,"datatype"),
        (GDS_XY,"xy")
        ],
    GDS_BOX_ELEM: [
        (GDS_BOX,{"element": "box"}),
        (GDS_ELFLAGS|GDS_OPTIONAL,"flags"),
        (GDS_PLEX|GDS_OPTIONAL,"plex"),
        (GDS_LAYER,"layer"),
        (GDS_BOXTYPE,"datatype"),
        (GDS_XY,"xy")
        ],
    GDS_STRANS_GROUP: [
        (GDS_STRANS,"strans"),
        (GDS_MAG|GDS_OPTIONAL,"mag"),
        (GDS_ANGLE|GDS_OPTIONAL,"angle"),
        ],
    GDS_TEXTBODY: [
        (GDS_TEXTTYPE,"datatype"),
        (GDS_PRESENTATION|GDS_OPTIONAL,"presentation"),
        (GDS_PATHTYPE|GDS_OPTIONAL,"pathtype"),
        (GDS_WIDTH|GDS_OPTIONAL,"width"),
        (GDS_STRANS_GROUP|GDS_OPTIONAL|GDS_INCLUDE,None),
        (GDS_XY,"xy"),
        (GDS_STRING,"text")
        ],
    GDS_PROPERTY: [
        (GDS_PROPATTR,"name"),
        (GDS_PROPVALUE,"value")
        ]
}

def gds_read(stream):
    def _gds_read_stream():
        while True:
            hdr = stream.read(4)
            if not hdr:
                break
            size, rec_id = struct.unpack(">HH", hdr)
            assert(rec_id in GDS_VALID_TAGS), f"Unknown GDS record encountered {hex(rec_id)}"
            data = stream.read(size - 4)
            yield rec_id, _gds_unpack_data(rec_id, data)
            if rec_id == GDS_ENDLIB:
                break
    rec_iter = iter(_gds_read_stream())
    rec_id, rec_val = (None, None)
    def _next_rec():
        nonlocal rec_id, rec_val
        try:
            rec_id, rec_val = next(rec_iter)
        except StopIteration:
            rec_id, rec_val = (None, None)
    _next_rec()
    def _try_tag(tag):
        if tag&GDS_NON_TOKEN:
            res = {}
            for _tag, name in _gds_parse_schema[tag]:
                if _tag == GDS_CHOICE:
                    v = None
                    for _tag in name:
                        v = _try_tag(_tag)
                        if v is not None:
                            _tag |= GDS_INCLUDE
                            break
                    name = None
                elif _tag&GDS_ARRAY:
                    v = list(_try_tag_array(_tag&GDS_TAGMASK))
                    if not v:
                        v = None
                else:
                    v = _try_tag(_tag&GDS_TAGMASK)
                if v is None:
                    if _tag&GDS_OPTIONAL:
                        continue
                    return None
                if _tag&GDS_INCLUDE:
                    res.update(v)
                elif name is not None:
                    if isinstance(name, str):
                        res[name] = v
                    else:
                        res.update(name)
            return res
        else:
            if rec_id == tag:
                v = rec_val
                _next_rec()
                return {} if v is None else v
    def _try_tag_array(tag):
        while True:
            v = _try_tag(tag)
            if v is None:
                break
            yield v
    return _try_tag(GDS_LIBRARY)

def gds_write(stream, gds):
    def _tag_val_match(tag, v):
        assert(tag in _gds_parse_schema),f"Wrong choice {hex(tag)} element in GDS schema"
        for _tag, name in _gds_parse_schema[tag]:
            if isinstance(name, str):
                if name in v:
                    continue
                if _tag&GDS_OPTIONAL:
                    continue
            elif isinstance(name, dict):
                for nm, _v in name.items():
                    if nm not in v or v[nm] != _v:
                        return False
        return True
    def _write_tag(tag, v):
        if tag&GDS_NON_TOKEN:
            for _tag, name in _gds_parse_schema[tag]:
                ok = False
                if _tag == GDS_CHOICE:
                    for _tag in name:
                        if _tag_val_match(_tag, v):
                            #print("match:",hex(_tag),v)
                            if _write_tag(_tag, v):
                                ok = True
                                break
                    assert(ok),f"GDS tree has wrong data for element {hex(tag)}: {v}"
                elif _tag&GDS_ARRAY:
                    if name in v:
                        for _v in v[name]:
                            if not _write_tag(_tag&GDS_TAGMASK, _v):
                                assert(False),f"GDS array {hex(_tag)} has wrong item value: {_v}"
                        ok = True
                elif _tag&GDS_INCLUDE:
                    ok =  _write_tag(_tag&GDS_TAGMASK, v)
                elif _tag&0xFF == GDS_NO_DATA:
                    ok =  _write_tag(_tag&GDS_TAGMASK, None)
                elif name in v:
                    ok =  _write_tag(_tag&GDS_TAGMASK, v[name])
                if not ok:
                    if _tag&GDS_OPTIONAL:
                        continue
                    return False
        else:
            data = _gds_pack_data(tag, v)
            hdr = struct.pack(">HH", len(data)+4, tag)
            stream.write(hdr)
            stream.write(data)
        return True
    return _write_tag(GDS_LIBRARY, gds)

def gds_new_timestamp():
    tm = list(time.localtime()[:6])
    return tm+tm

def gds_create(lib_name, units = None):
    if not units:
        units = [0.001, 1e-09]
    return {"version": 5,
            "timestamp": gds_new_timestamp(),
            "name": lib_name,
            "units": units,
            "structures": []
           }

def gds_find_structure(gds, name):
    for st in gds["structures"]:
        if st["name"] == name:
            return st

def gds_get_structures(gds):
    return {st["name"]:st for st in gds["structures"]}

def gds_get_structures_names(gds):
    return [st["name"] for st in gds["structures"]]

def gds_create_structure(name):
    return {"timestamp": gds_new_timestamp(),
            "name": name,
            "elements": []}

def gds_flip_rotate_pt(pt, flip, angle):
    if flip:
        pt = (pt[0], -pt[1])
    if not angle:
        return pt
    elif angle == 90:
        return (-pt[1], pt[0])
    elif angle == 180:
        return (-pt[0], -pt[1])
    elif angle == 270:
        return (pt[1], -pt[0])
    angle = math.radians(angle)
    sn = math.sin(angle)
    cs = math.cos(angle)
    return (int(pt[0]*cs - pt[1]*sn), int(pt[0]*sn + pt[1]*cs))


def gds_get_structures_names_with_deps(gds, st_names):
    names = set()
    def _srefs(st):
        for el in st["elements"]:
            if el["element"] == "sref":
                yield el
    def _populate(name):
        names.add(name)
        st = gds_find_structure(gds, name)
        assert(st), f"GDS referencing non existing structure '{name}"
        for el in _srefs(st):
            el_name = el["name"]
            if el_name in names:
                continue
            _populate(el_name)
    for name in st_names:
        _populate(name)
    # dependancy sort
    ordered_names = sorted(names)
    i = 0
    nerr = 0
    while i<len(ordered_names):
        name = ordered_names[i]
        st = gds_find_structure(gds, name)
        if any(map(lambda el: el["name"] in  names, _srefs(st))):
            ordered_names.pop(i)
            ordered_names.append(name)
            nerr += 1
            assert(nerr < len(names)), "GDS structures have elements with cross dependancies"
        else:
            names.discard(name)
            i += 1
            nerr = 0
    return ordered_names



import math

pi_half = math.pi / 2

class Vector2D:
    def __init__(self, x, y):
        self.x = x
        self.y = y

    def dot(self, v):
        return self.x*v.x + self.y*v.y

    def size(self):
        return math.sqrt(self.size2())

    def size2(self):
        return self.x**2 + self.y**2

    def sized(self, sz):
        _sz = self.size()
        return Vector2D(self.x * sz / _sz, self.y * sz / _sz)

    def rotated(self, angle):
        # cos(a+b)=cos(a)*cos(b)-sin(a)*sin(b)
        # sin(a+b)=sin(a)*cos(b)+cos(a)*sin(b)
        v = Vector2D(math.cos(angle), -math.sin(angle))
        return Vector2D(self.dot(v), self.normal().dot(v))

    def normal(self):
        return Vector2D(self.y, -self.x)

    def tangent(self):
        return self.y / self.x

    def angle(self):
        if self.x == 0:
            return -pi_half if self.y <0 else pi_half
        if self.y == 0:
            return math.pi if self.x < 0 else 0
        angle = math.atan(self.tangent())
        if self.x < 0:
            angle += -math.pi if self.y < 0 else math.pi
        return angle

    def winding_angle(self, start, end):
        if start == end:
            return 0
        # cos(A-B)=cos(A)*cos(B)+sin(a)*sin(B)
        # sin(A-B)=sin(A)*cos(B)-cos(a)*sin(B)
        v1 = start - self
        v2 = end - self
        v = Vector2D(v1.dot(v2),v1.dot(v2.normal())) # angle(v)=angle(v2)-angle(v1), size(v)=size(v1)*size(v2)
        return v.angle()

    def __add__(self, v):
        if isinstance(v, Vector2D):
            x = v.x
            y = v.y
        else:
            x, y = v
        return Vector2D(self.x + x, self.y + y)

    def __radd__(self, v):
        return self + v

    def __sub__(self, v):
        if isinstance(v, Vector2D):
            x = v.x
            y = v.y
        else:
            x, y = v
        return Vector2D(self.x - x, self.y - y)

    def __rsub__(self, v):
        if isinstance(v, Vector2D):
            x = v.x
            y = v.y
        else:
            x, y = v
        return Vector2D(x - self.x, y - self.y)

    def inside(self, pts):
        pts = iter(pts)
        pstart = next(pts)
        prev = pstart
        wangle = 0
        while True:
            try:
                pt = next(pts)
                wangle += self.winding_angle(prev, pt)
                prev = pt
            except StopIteration:
                break
        if prev != pstart:
            wangle += self.winding_angle(prev, pstart)
        return abs(wangle) > 0.1 # should be zero (up to accuracy) if outside
